#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the plotting of chemical properties using pyvista.
"""

import numpy as np
import pyvista
from rgmol.objects import *
import rgmol.molecular_calculations
from ._functions import *

###########################
## function for subplots ##
###########################


def _subplotting(number_molecules):
    """
    _subplotting(number_molecules)

    Function that finds the proper number of subplots for group of molecules plotting

    Parameters
    ----------
        number_molecules : int
            The number of molecules

    Returns
    -------
        tuple
            The shape of the plotter
    """

    if number_molecules < 4:
        return (1,number_molecules)
    elif number_molecules < 7:
        return (2,int(round(number_molecules/2)))
    elif number_molecules < 10:
        print("Lots of molecules")
        return (3,int(round(number_molecules/3)))
    else:
        print("Too much molecules, the graph will still be plotted, but it will not be good")
        return (3,int(round(number_molecules/3)))

def _get_subplot(number_molecules,index):
    """
    _get_subplot(number_molecules,index)

    Gets the 2d subplot index from the index and the number of molecules

    Parameters
    ----------
        number_molecules : int
            The number of molecules
        index : int
            The index of the molecule

    Returns
    -------
        tuple
            The position of the index on the subplots
    """

    if number_molecules < 4:
        return (0,index)
    elif number_molecules < 7:
        frac = int(round(number_molecules/2))
        return (index//frac,index%frac)
    else:
        frac = int(round(number_molecules/3))
        return (index//frac,index%frac)


class _subplot:
    """small class used for sliders in subplots"""
    def __init__(self,indexes,plotter,func,molecule):
        self.indexes = indexes
        self.plotter = plotter
        self.func = func
        self.molecule = molecule

    def __call__(self, value):
        self.func(value,self.indexes,self.molecule)


class _subplot_isodensity:
    """small class used for sliders in subplots"""
    def __init__(self,indexes,plotter,cutoff,func,molecule,value):
        self.indexes = indexes
        self.plotter = plotter
        self.cutoff = cutoff
        self.func = func
        self.molecule = molecule
        self.value = value


    def __call__(self, value):
        self.value = value
        self.func(self.value,self.indexes,self.molecule,self.cutoff)
    def call_cutoff(self,cutoff):
        self.cutoff = cutoff
        self.func(self.value,self.indexes,self.molecule,self.cutoff)



#####################################################
## Adding Plotting Methods for Groud of Molecules  ##
#####################################################




def plot_radius(self,opacity=1,factor=.4,show_bonds=True,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_radius(opacity=1,factor=.4,show_bonds=True,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the radius of each molecules

    Parameters
    ----------
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        show_bonds : bool, optional
            Choose to show the bonds between the atoms or not. By default True
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """
    number_molecules = len(self.molecules)
    shape = _subplotting(number_molecules)
    plotter = pyvista.Plotter(shape=shape,border=True)
    for index_molecule in range(number_molecules):
        index = _get_subplot(number_molecules,index_molecule)
        plotter.subplot(index[0],index[1])

        molecule = self.molecules[index_molecule]
        for atom_x in molecule.atoms:
            atom_x.plot(plotter,opacity=opacity,factor=factor)
        if show_bonds:
            bonds_plotting(plotter,molecule.bonds,molecule.list_property("pos"),molecule.list_property("radius"),factor=factor)

        light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
        plotter.add_light(light)
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return


def plot_property(self,plotted_property,opacity=.4,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_property(plotted_property,opacity=.4,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot a property for each molecule

    Parameters
    ----------
        plotted_property : float, optional
            The property to be plotted. By default the radius
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    number_molecules = len(self.molecules)
    shape = _subplotting(number_molecules)
    plotter = pyvista.Plotter(shape=shape,border=True)
    for index_molecule in range(number_molecules):
        index = _get_subplot(number_molecules,index_molecule)
        plotter.subplot(index[0],index[1])

        molecule = self.molecules[index_molecule]

        X = molecule.properties[plotted_property]

        if with_radius:
            molecule.plot(plotter,factor=factor_radius,opacity=opacity_radius)
        molecule.plot_vector(plotter,X,opacity=opacity,factor=factor)
        light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
        plotter.add_light(light)

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return


def plot_diagonalized_condensed_kernel(self,kernel,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_diagonalized_condensed_kernel(kernel,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000))

    Diagonalize and plot a condensed kernel for each molecule. One can navigate through the eigenmodes using a slider for each molecule.

    Parameters
    ----------
        kernel : str
            The kernel to be diagonalized and plotted
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """


    number_molecules = len(self.molecules)
    shape = _subplotting(number_molecules)
    plotter = pyvista.Plotter(shape=shape,border=True)

    def create_mesh_diagonalized_kernel(value,indexes,molecule):
        plotter.subplot(indexes[0],indexes[1])
        vector_number = int(round(value))
        molecule.plot_vector(plotter,XV[:,vector_number-1],opacity=opacity,factor=factor)
        plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(Xvp[vector_number-1]),name="eigenvalue")

    list_subpl = [_subplot(_get_subplot(number_molecules,index_molecule),plotter,create_mesh_diagonalized_kernel,self.molecules[index_molecule]) for index_molecule in range(number_molecules)]

    for index_molecule in range(number_molecules):
        index = _get_subplot(number_molecules,index_molecule)
        plotter.subplot(index[0],index[1])

        molecule = self.molecules[index_molecule]

        X = molecule.properties[kernel]
        Xvp,XV = np.linalg.eigh(X)

        if with_radius:
            molecule.plot(plotter,factor=factor_radius,opacity=opacity_radius)

        light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
        plotter.add_light(light)
        plotter.add_slider_widget(list_subpl[index_molecule], [1, len(XV)],value=1,title="Eigenvector", fmt="%1.0f")
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return


def plot_isodensity(self,plotted_isodensity="cube",opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_isodensity(plotted_isodensity="cube",opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot an isodensity for each molecule

    Parameters
    ----------
        plotted_isodensity : str, optional
            The isodensity to be plotted. By default "cube"
        cutoff : float, optional
            The cutoff of the isodensity plot. By default .2
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    number_molecules = len(self.molecules)
    shape = _subplotting(number_molecules)
    plotter = pyvista.Plotter(shape=shape,border=True)
    for index_molecule in range(number_molecules):
        index = _get_subplot(number_molecules,index_molecule)
        plotter.subplot(index[0],index[1])

        molecule = self.molecules[index_molecule]
        if with_radius:
            molecule.plot(plotter,factor=factor_radius,opacity=opacity_radius,show_bonds=True)
        plot_cube(plotter,molecule.properties["voxel_origin"],molecule.properties["voxel_matrix"],molecule.properties["cube"],opacity=opacity,factor=factor,cutoff=cutoff)

        light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
        plotter.add_light(light)
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return





def plot_AO(self,grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_AO(grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the Atomic Orbitals of each molecule
    The Atomic Orbitals will be calculated on the grid that will be defined by the number of grid points and around the molecule. The delta defines the length to be added to the extremities of the position of the atoms.
    The order of the Atomic Orbitals is defined in the molden file

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    number_molecules = len(self.molecules)
    shape = _subplotting(number_molecules)
    plotter = pyvista.Plotter(shape=shape,border=True)

    def create_mesh_AO(value,indexes,molecule,cutoff):
        plotter.subplot(indexes[0],indexes[1])
        AO_number = int(round(value))
        plot_cube(plotter,molecule.properties["voxel_origin"],molecule.properties["voxel_matrix"],molecule.properties["AO_calculated"][int(round(value))-1],opacity=opacity,factor=factor,cutoff=cutoff,add_name=str(shape[1]*indexes[0]+indexes[1]))

    list_subpl = [_subplot_isodensity(_get_subplot(number_molecules,index_molecule),plotter,cutoff,create_mesh_AO,self.molecules[index_molecule],1) for index_molecule in range(number_molecules)]

    for index_molecule in range(number_molecules):
        index = _get_subplot(number_molecules,index_molecule)
        plotter.subplot(index[0],index[1])

        molecule = self.molecules[index_molecule]

        if not "AO_calculated" in molecule.properties:
            rgmol.molecular_calculations.calculate_AO(molecule,grid_points,delta=delta)

        if with_radius:
            molecule.plot(plotter,factor=factor_radius,opacity=opacity_radius)

        light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
        plotter.add_light(light)
        plotter.add_slider_widget(list_subpl[index_molecule], [1, len(molecule.properties["AO_calculated"])],value=1,title="Number", fmt="%1.0f")
        plotter.add_slider_widget(list_subpl[index_molecule].call_cutoff, [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)
    return





def plot_MO(self,grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_MO(grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the Molecular Orbitals of each molecule
    The Molecular Orbitals will be calculated on the grid that will be defined by the number of grid points and around the molecule. The delta defines the length to be added to the extremities of the position of the atoms.

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    number_molecules = len(self.molecules)
    shape = _subplotting(number_molecules)
    plotter = pyvista.Plotter(shape=shape,border=True)

    def create_mesh_MO(value,indexes,molecule,cutoff):
        plotter.subplot(indexes[0],indexes[1])
        MO_number = int(round(value))
        MO_calculated = molecule.calculate_MO_chosen(MO_number-1,grid_points,delta=delta)

        plot_cube(plotter,molecule.properties["voxel_origin"],molecule.properties["voxel_matrix"],MO_calculated,opacity=opacity,factor=factor,cutoff=cutoff,add_name=str(shape[1]*indexes[0]+indexes[1]))
        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(molecule.properties["MO_energy"][MO_number-1]),name="mo energy")
        print_occupancy(plotter,molecule.properties["MO_occupancy"],MO_number,divy=shape[0])


    list_subpl = [_subplot_isodensity(_get_subplot(number_molecules,index_molecule),plotter,cutoff,create_mesh_MO,self.molecules[index_molecule],1) for index_molecule in range(number_molecules)]

    for index_molecule in range(number_molecules):
        index = _get_subplot(number_molecules,index_molecule)
        plotter.subplot(index[0],index[1])

        molecule = self.molecules[index_molecule]

        if not "MO_calculated" in molecule.properties:
            molecule.properties["MO_calculated"] = [[] for k in range(len(molecule.properties["MO_list"]))]

        if with_radius:
            molecule.plot(plotter,factor=factor_radius,opacity=opacity_radius)


        light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
        plotter.add_light(light)
        plotter.add_slider_widget(list_subpl[index_molecule], [1, len(molecule.properties["MO_calculated"])],value=1,title="Number", fmt="%1.0f")
        plotter.add_slider_widget(list_subpl[index_molecule].call_cutoff, [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)
    return




def plot_transition_density(self,grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_transition_density(grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))


    Plot the Transition Densities of each molecule.
    All the AO and the MO will be calculated on the grid if they were not calculated.
    The grid is defined by the number of grid points and around the molecule. The delta defines the length to be added to the extremities of the position of the atoms.

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function

    Notes
    -----
    The transition densities are defined as :

    | :math:`\\rho_0^k = \\sum_i c_i (Occ_i(r) * Virt_i(r))`
    | With the sum being on all the transitions of the excitation, :math:`Occ_i(r)` and :math:`Virt_i(r)` being respectively the occupied and the virtual molecular orbitals considered in the transition, and :math:`c_i` the coefficient of the transition.

    Examples
    --------
    For a TD-DFT calculation output :

    STATE  1:  E=   0.148431 au      4.039 eV
       | 18a ->  20a  :     0.5000 (c= 0.7071)
       | 19a ->  21a  :     0.5000 (c= -0.7071)

    The transition density will be :
        | :math:`\\rho_0^1 =  c_1 (\\psi_{18}(r) * \\psi_{20}(r)) + c_2 (\\psi_{19}(r) * \\psi_{21}(r))`
        | with :math:`c_1 = 0.7071` and :math:`c_2 = -0.7071`
    """


    number_molecules = len(self.molecules)
    shape = _subplotting(number_molecules)
    plotter = pyvista.Plotter(shape=shape,border=True)

    def create_mesh_transition_density(value,indexes,molecule,cutoff):
        plotter.subplot(indexes[0],indexes[1])
        transition_number = int(round(value))
        transition_density_calculated = molecule.calculate_chosen_transition_density(transition_number-1,grid_points,delta=delta)

        plot_cube(plotter,molecule.properties["voxel_origin"],molecule.properties["voxel_matrix"],transition_density_calculated,opacity=opacity,factor=factor,cutoff=cutoff,add_name=str(shape[1]*indexes[0]+indexes[1]))

        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(molecule.properties["transition_energy"][transition_number-1]),name="transition energy")

    list_subpl = [_subplot_isodensity(_get_subplot(number_molecules,index_molecule),plotter,cutoff,create_mesh_transition_density,self.molecules[index_molecule],1) for index_molecule in range(number_molecules)]

    for index_molecule in range(number_molecules):
        index = _get_subplot(number_molecules,index_molecule)
        plotter.subplot(index[0],index[1])

        molecule = self.molecules[index_molecule]
        #Initialize MO list
        if not "MO_calculated" in molecule.properties:
            molecule.properties["MO_calculated"] = [[] for k in range(len(molecule.properties["MO_list"]))]

        #Initialize transition density list
        if not "transition_density_list" in molecule.properties:
            molecule.properties["transition_density_list"] = [[] for k in range(len(molecule.properties["transition_list"]))]

        if with_radius:
            molecule.plot(plotter,factor=factor_radius,opacity=opacity_radius)

        light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
        plotter.add_light(light)
        plotter.add_slider_widget(list_subpl[index_molecule], [1, len(molecule.properties["transition_density_list"])],value=1,title="Number", fmt="%1.0f")
        plotter.add_slider_widget(list_subpl[index_molecule].call_cutoff, [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)





def plot_diagonalized_kernel(self,kernel,method="only eigenmodes",plotting_method="isodensity",number_eigenvectors=20,grid_points=(20,20,20),delta=3,number_isodensities=10 ,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_diagonalized_kernel(kernel,method="only eigenmodes",plotting_method="isodensity",number_eigenvectors=20,grid_points=(20,20,20),delta=3,number_isodensities=10 ,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Calculate and diagonalize a kernel. Only the linear response function is implemented for now.
    Only the first number_eigenvectors will be computed to limit the calculation time.
    The grid is defined by the number of grid points and around the molecule. The delta defines the length to be added to the extremities of the position of the atoms.

    Parameters
    ----------
        kernel : str
            The kernel to be diagonalized and plotted
        method : str, optional
            The method of calculation of the kernel, only eigenmodes by default. More information can be found in the notes.
        plotting_method : str, optional
            The method used for plotting. By default isodensity. The other methods are : "multiple isodensities", "volume"
        number_eigenvectors : int, optional
            Only used in the methods "total" and "partial". The number of eigenvectors to be computed
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        cutoff : float, optional
            The initial cutoff of the isodensity plot for the isodensity plotting method. By default .2
        number_isodensities : int, optional
            The number of isodensities to be plotted if the method used is multiple isodensities
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function


    Notes
    -----
        Because the kernels are 6-dimensional, they scale up drastically in terms of memory used.
        If one only wants to look at the eigenmodes, the "only eigenmodes" method is just that. It computes the eigenmodes without computing the total kernel. More info can be found :doc:`here<../orbitals/calculate_eigenmodes_linear_response_function>`.
        Otherwise, the partial method has been implemented which allows to remove the part of the space where the transition densities are almost zero. For each transition density, the space is sorted and the lower dense part that makes up to less than 1% is removed. In practice this removes as much as 90% of the space. More details on this method can be found :doc:`here<../orbitals/diagonalize_kernel>`.
        The last method is "total" which is just the calculation of the linear response on the whole space

    """

    number_molecules = len(self.molecules)
    shape = _subplotting(number_molecules)
    plotter = pyvista.Plotter(shape=shape,border=True)


    if plotting_method == "isodensity":
        def create_mesh_diagonalized_kernel(value,indexes,molecule,cutoff):
            plotter.subplot(indexes[0],indexes[1])
            vector_number = int(round(value))
            eigenvectors = molecule.properties["linear_response_eigenvectors"]
            eigenvalues = molecule.properties["linear_response_eigenvalues"]
            contrib_eigenvectors = molecule.properties["contibution_linear_response_eigenvectors"]
            transition_list = molecule.properties["transition_list"]
            transition_factor_list = molecule.properties["transition_factor_list"]

            plot_cube(plotter,molecule.properties["voxel_origin"],molecule.properties["voxel_matrix"],eigenvectors[vector_number-1],opacity=opacity,factor=factor,cutoff=cutoff,add_name=str(shape[1]*indexes[0]+indexes[1]))
            plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue",font_size=18/shape[1])
            if method == "only eigenmodes":
                print_contribution_transition_density(plotter,vector_number,contrib_eigenvectors,transition_list,transition_factor_list,divy=shape[0])

    elif plotting_method == "multiple isodensities":
        def create_mesh_diagonalized_kernel(value,indexes,molecule,cutoff):
            plotter.subplot(indexes[0],indexes[1])
            vector_number = int(round(value))
            eigenvectors = molecule.properties["linear_response_eigenvectors"]
            eigenvalues = molecule.properties["linear_response_eigenvalues"]
            contrib_eigenvectors = molecule.properties["contibution_linear_response_eigenvectors"]
            transition_factor_list = molecule.properties["transition_factor_list"]
            transition_list = molecule.properties["transition_list"]

            plot_cube_multiple_isodensities(plotter,molecule.properties["voxel_origin"],molecule.properties["voxel_matrix"],eigenvectors[vector_number-1],factor=factor,add_name=str(shape[1]*indexes[0]+indexes[1]))
            plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue",font_size=18/shape[1])

            if method == "only eigenmodes":
                print_contribution_transition_density(plotter,vector_number,contrib_eigenvectors,transition_list,transition_factor_list,divy=shape[0])

    elif plotting_method == "volume":
        def create_mesh_diagonalized_kernel(value,indexes,molecule,cutoff):
            plotter.subplot(indexes[0],indexes[1])
            vector_number = int(round(value))
            eigenvectors = molecule.properties["linear_response_eigenvectors"]
            eigenvalues = molecule.properties["linear_response_eigenvalues"]
            contrib_eigenvectors = molecule.properties["contibution_linear_response_eigenvectors"]
            transition_list = molecule.properties["transition_list"]
            transition_factor_list = molecule.properties["transition_factor_list"]


            plot_cube_volume(plotter,molecule.properties["voxel_origin"],molecule.properties["voxel_matrix"],eigenvectors[vector_number-1],factor=factor,add_name=str(shape[1]*indexes[0]+indexes[1]))
            plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue",font_size=18/shape[1])

            if method == "only eigenmodes":
                print_contribution_transition_density(plotter,vector_number,contrib_eigenvectors,transition_list,transition_factor_list,divy=shape[0])




    list_subpl = [_subplot_isodensity(_get_subplot(number_molecules,index_molecule),plotter,cutoff,create_mesh_diagonalized_kernel,self.molecules[index_molecule],1) for index_molecule in range(number_molecules)]

    for index_molecule in range(number_molecules):
        index = _get_subplot(number_molecules,index_molecule)
        plotter.subplot(index[0],index[1])

        molecule = self.molecules[index_molecule]
        if kernel != "linear_response_function":
            raise ValueError("Only linear response function implemented for now")

        if method == "only eigenmodes":
            molecule.calculate_eigenmodes_linear_response_function(grid_points,delta=delta)
        else:
            molecule.diagonalize_kernel(kernel,number_eigenvectors,grid_points,method=method,delta=delta)

        eigenvalues = molecule.properties["linear_response_eigenvalues"]

        if with_radius:
            molecule.plot(plotter,factor=factor_radius,opacity=opacity_radius)

        light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
        plotter.add_light(light)
        plotter.add_slider_widget(list_subpl[index_molecule], [1, len(eigenvalues)],value=1,title="Eigenvector", fmt="%1.0f")
        if plotting_method == "isodensity":
            plotter.add_slider_widget(list_subpl[index_molecule].call_cutoff, [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)




group_molecules.plot_radius = plot_radius
group_molecules.plot_property = plot_property
group_molecules.plot_diagonalized_condensed_kernel = plot_diagonalized_condensed_kernel
group_molecules.plot_isodensity = plot_isodensity
group_molecules.plot_AO = plot_AO
group_molecules.plot_MO = plot_MO
group_molecules.plot_transition_density = plot_transition_density
group_molecules.plot_diagonalized_kernel = plot_diagonalized_kernel




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds writing functions
"""

import numpy as np
import scipy as sp
from rgmol.objects import *
import os

def _write_vector_for_cube(file,vector):
    """Writes a vector in a file"""
    for elem in vector:
        long = len(str(abs(int(elem))))
        file.write(" "*(4-long))
        file.write("{:9.6f}".format(elem))

def _write_data_for_cube(file,vector):
    """Writes a vector in a file"""
    for elem in vector:
        m = int(np.sign(elem)/2-1/2)
        file.write(" "*(3+m))
        file.write("{:9.5e}".format(elem))

def _get_file_location(file):
    """gets the location of the file of the output of the molecule"""
    file_splitted = file.split("//")[:-1]
    file_location = ""

    for f in file_splitted:
        file_location += f
        file_location +="//"

    return file_location



def write_cube(self,cube,file_name,description="cube"):
    """
    write_cube(cube,file_name,description="cube")

    Writes a cube file of a 3D vector inside a folder called rgmol which will be located inside the folder containing the molecule.

    Parameters
    ----------
        cube : ndarray
            The 3D vector to be stored in the cube file
        file_name : str
            The name of the file
        description : str
            Description of the file inside the cube file

    Returns
    -------
        None
            The cube file is written
    """

    file_location = _get_file_location(self.file)

    if not "rgmol" in os.listdir(file_location):
        os.mkdir(file_location+"rgmol")

    with open(file_location+"rgmol//"+file_name,"w") as file:
        file.write("Cube data generated by rgmol\n")
        file.write(description+"\n")

        #Number atoms and voxel origin
        num_atoms = str(len(self.atoms))
        voxel_origin = self.properties["voxel_origin"]

        file.write(" "*(5-len(num_atoms)) +num_atoms)
        _write_vector_for_cube(file,voxel_origin)

        file.write("\n")

        #Number of points and voxel matrix
        voxel_matrix = self.properties["voxel_matrix"]
        grid_points = self.properties["grid_points"]

        grid_points_0 = str(grid_points[0])
        file.write(" "*(5-len(grid_points_0)) + grid_points_0)
        _write_vector_for_cube(file,voxel_matrix[0])
        file.write("\n")
        grid_points_1 = str(grid_points[1])
        file.write(" "*(5-len(grid_points_1)) + grid_points_1)
        _write_vector_for_cube(file,voxel_matrix[1])
        file.write("\n")
        grid_points_2 = str(grid_points[2])
        file.write(" "*(5-len(grid_points_2)) + grid_points_2)
        _write_vector_for_cube(file,voxel_matrix[2])
        file.write("\n")

        #Write atom positions
        for atom in self.atoms:
            atomic_number = atom.atomic_number
            file.write(" "*(5-len(str(atomic_number)))+str(atomic_number))
            file.write(" "*(4-len(str(atomic_number)))+"{:9.6f}".format(atomic_number))
            _write_vector_for_cube(file,atom.pos)
            file.write("\n")

        #Write cube
        for vec_x in cube:
            for vec_y in vec_x:
                line_length = grid_points[2]
                count = 0
                while count < line_length:
                    _write_data_for_cube(file,vec_y[count:count+6])
                    file.write('\n')
                    count+=6




molecule.write_cube = write_cube







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script contains the functions used to extract data from molden files.
"""


import codecs
import numpy as np
from rgmol.objects import *
from rgmol.general_function import find_bonds


##########################
## Extraction functions ##
##########################


def _extract_molden_file(file):
    """
    _extract_molden_file(file)

    Extracts the data from a molden file

    Parameters
    ----------
        file : str

    Returns
    -------
        atom_names : list
        atom_position : ndarray
        AO_list : list
        AO_type_list : list
        MO_list : list
        MO_energy : list
    """

    flag_atoms = 0
    flag_gto = 0
    flag_mo = 0
    flag_change = 0
    flag_mo_lines = 4
    atom_names = []
    atom_position = []

    AO_list = []
    AO_type_list = []

    MO_list = []
    MO_energy = []
    spin = []
    MO_occupancy = []
    D = np.arange(154)

    for line in codecs.open(file, 'r',encoding="utf-8"):
        lsplit = line.split()
        if "[Atoms]" in line:
            flag_atoms = 1
            flag_gto = 0
            flag_mo = 0
            flag_change = 1
        if "[GTO]" in line:
            flag_atoms = 0
            flag_gto = 1
            flag_mo = 0
            flag_change = 1

            flag_gto_atom = 1

        if '[MO]' in line:
            flag_atoms = 0
            flag_gto = 0
            flag_mo = 1
            flag_change = 1

        if "[5D]" in line or "[7F]" in line or "[9G]" in line :
            flag_atoms = 0
            flag_gto = 0
            flag_mo = 1
            flag_change = 1


        if flag_change:#Skip the line
            flag_change = 0

        elif flag_atoms:
            atom_names.append(lsplit[0])
            atom_position.append([float(lsplit[3]),float(lsplit[4]),float(lsplit[5])])

        elif flag_gto:
            if flag_gto_atom:#New atom
                AO_atom = []
                AO_atom_orbital_type = []
                flag_gto_atom = 0
                flag_gto_orb_num = 1

            elif flag_gto_orb_num:#New orbital name
                if len(line) <= 2:#Empty line meaning change of atom
                    flag_gto_orb_num = 0
                    flag_gto_atom = 1
                    AO_list.append(AO_atom)
                    AO_type_list.append(AO_atom_orbital_type)
                else:
                    AO_orbital = []
                    AO_atom_orbital_type.append(lsplit[0])
                    num_orb = int(lsplit[1])
                    flag_gto_orb_num = 0
                    flag_gto_orb = 1

            elif flag_gto_orb:#New orbital value
                AO_orbital.append([float(lsplit[0]),float(lsplit[1])])
                num_orb -= 1

                if num_orb == 0:#Last value
                    AO_atom.append(AO_orbital)

                    flag_gto_orb = 0
                    flag_gto_orb_num = 1

        elif flag_mo:
            if len(line) <= 2:#Last line for molecular orbitals
                MO_list.append(AO_contribution)

            #The first 4 lines contain details
            elif flag_mo_lines <= 4 and flag_mo_lines > 0:
                if lsplit[0]=="Ene=":
                    MO_energy.append(float(lsplit[1]))

                if lsplit[0] == "Spin=" and lsplit[1] == "Beta":
                    raise ValueError("Unrestricted calculations not currently implemented")
                    # spin.append(lsplit[1])
                if lsplit[0] == "Occup=":
                    MO_occupancy.append(float(lsplit[1]))
                AO_contribution = []

                flag_mo_lines -= 1

            #The contribution of each AO
            else:
                if lsplit[0]=="Sym=" or lsplit[0] =="Ene=":#New MO
                    if lsplit[0]=="Ene=":
                        MO_energy.append(float(lsplit[1]))

                    if lsplit[0] == "Spin=" and lsplit[1] == "Beta":
                        raise ValueError("Unrestricted calculations not currently implemented")
                        # spin.append(lsplit[1])
                    if lsplit[0] == "Occup=":
                        MO_occupancy.append(float(lsplit[1]))

                    flag_mo_lines = 3
                    MO_list.append(AO_contribution)
                else:
                    AO_contribution.append(float(lsplit[1]))
    if AO_contribution != MO_list[-1]:
        MO_list.append(AO_contribution)
    return atom_names,atom_position,AO_list,AO_type_list,MO_list,MO_energy,MO_occupancy


def extract(file,do_find_bonds=0):
    """
    extract(file,do_find_bonds=0)

    Extract from molden input and create molecule object.
    As the bonds are not defined in a molden file, one can use do_find_bonds to
    use an algorithm that tries to find bonds. It is still in WIP

    Parameters
    ----------
        file : str
        do_find_bonds : bool, optional
            if one wants to use an algorithm to find the bonds (WIP)

    Returns
    -------
        mol : molecule
    """

    atom_names,atom_position,AO_list,AO_type_list,MO_list,MO_energy,MO_occupancy = _extract_molden_file(file)


    list_atoms = []
    nicknaming = 0
    for prop in zip(atom_names,atom_position):
        atom_x = atom(prop[0],prop[1],nickname=str(nicknaming))
        list_atoms.append(atom_x)
        nicknaming+=1

    mol = molecule(list_atoms,[],file=file,properties={"AO_list":AO_list,"AO_type_list":AO_type_list,"MO_list":MO_list,"MO_energy":MO_energy,"MO_occupancy":MO_occupancy})

    if do_find_bonds:
        mol.bonds = find_bonds(mol)

    return mol





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the plotting of chemical properties using pyvista.
"""

import numpy as np
import pyvista
from rgmol.objects import *
import rgmol.molecular_calculations
from ._functions import *

###############################
## Defines class for sliders ##
###############################

class _slider:
    def __init__(self,func,value,cutoff):
        self.func = func
        self.value = value
        self.cutoff = cutoff

    def __call__(self, slider, value):
        if slider == "cutoff":
            self.cutoff = value
        else:
            self.value = value
        self.func(self.value,self.cutoff)

########################################
## Adding Plotting Methods for Atoms  ##
########################################



def plot(self,plotter,plotted_property="radius",opacity=1,factor=1):
    """
    plot(plotter,plotted_property="radius",opacity=1,factor=1)

    Plot a property of the atom on the plotter using pyvista

    Parameters :
    ------------
        plotter : pyvista.plotter
            The plotter object from pyvita on which the atom will be plotted. It can be easily defined using plotter = pyvista.Plotter()
        plotted_property : string, optional
            The property to be plotted. By default the radius is plotted.
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1

    Returns :
    ---------
        None
            The atom is plotted on the plotter object
    """
    plot_atom(plotter,self,plotted_property=plotted_property,opacity=opacity,factor=factor)



def plot_vector(self,plotter,vector,opacity=1,factor=1):
    """
    plot_vector(plotter,plotted_property="radius",opacity=1,factor=1)

    Plot a value of a vector on the position of the atom using pyvista

    Parameters :
    ------------
        plotter : pyvista.plotter
            The plotter object from pyvita on which the atom will be plotted. It can be easily defined using plotter = pyvista.Plotter()
        vector : float
            The value to be plotted
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1

    Returns :
    ---------
        None
            The atom is plotted on the plotter object
    """
    plot_vector_atom(plotter,self,vector,opacity=opacity,factor=factor)


atom.plot = plot
atom.plot_vector = plot_vector


############################################
## Adding Plotting Methods for Molecules  ##
############################################



def plot(self,plotter,plotted_property="radius",opacity=1,factor=1,show_bonds=True):
    """
    plot(plotter,plotted_property="radius",opacity=1,factor=1,show_bonds=True)

    Plot a property for the entire molecule

    Parameters
    ----------
        plotter : pyvista.plotter
            The plotter object from pyvita on which the atom will be plotted. It can be easily defined using plotter = pyvista.Plotter()
        plotted_property : float, optional
            The property to be plotted. By default the radius
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        show_bonds : bool, optional
            Chose to show the bonds between the atoms or not. By default True

    Returns
    -------
        None
            The atom is plotted on the plotter object
    """
    for atom_x in self.atoms:
        atom_x.plot(plotter,plotted_property=plotted_property,opacity=opacity,factor=factor)
    if show_bonds:
        bonds_plotting(plotter,self.bonds,self.list_property("pos"),self.list_property(plotted_property),factor=factor)
    return



def plot_vector(self,plotter,vector,opacity=1,factor=1):
    """
    plot_vector(plotter,vector,opacity=1,factor=1,show_bonds=True)

    Plot a vector on each atom for the entire molecule

    Parameters
    ----------
        plotter : pyvista.plotter
            The plotter object from pyvita on which the atom will be plotted. It can be easily defined using plotter = pyvista.Plotter()
        vector : list or ndarray
            The vector to be plotted. The length should be equal to the number of atoms.
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        show_bonds : bool, optional
            Choose to show the bonds between the atoms or not. By default True

    Returns
    -------
        None
            The atom is plotted on the plotter object
    """
    for atom_x in range(len(self.atoms)):
        self.atoms[atom_x].plot_vector(plotter,vector[atom_x],opacity=opacity,factor=factor)
    return


def plot_radius(self,opacity=1,factor=.4,show_bonds=True,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_radius(opacity=1,factor=.4,show_bonds=True,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the radius of the entire molecule

    Parameters
    ----------
    opacity : float, optional
        The opacity of the plot. By default equals to 1
    factor : float, optional
        The factor by which the plotted_property will be multiplied. By default equals to 1
    show_bonds : bool, optional
        Choose to show the bonds between the atoms or not. By default True
    screenshot_button : bool, optional
        Adds a screenshot button. True by default
    window_size_screenshot : tuple, optional
        The size of the screenshots. By default (1000,1000)

    Returns
    -------
    None
        The plotter should display when using this function
    """
    plotter = pyvista.Plotter()
    for atom_x in self.atoms:
        atom_x.plot(plotter,opacity=opacity,factor=factor)
    if show_bonds:
        bonds_plotting(plotter,self.bonds,self.list_property("pos"),self.list_property("radius"),factor=factor)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return


def plot_property(self,plotted_property,opacity=.4,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_property(plotted_property,opacity=.4,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot a property for the entire molecule

    Parameters
    ----------
        plotted_property : float, optional
            The property to be plotted. By default the radius
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """
    X = self.properties[plotted_property]
    plotter = pyvista.Plotter()

    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
    self.plot_vector(plotter,X,opacity=opacity,factor=factor)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return


def plot_diagonalized_condensed_kernel(self,kernel,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_diagonalized_condensed_kernel(kernel,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,screenshot_button=True,window_size_screenshot=(1000,1000))

    Diagonalize and plot a condensed kernel for a molecule. One can navigate through the eigenmodes using a slider.

    Parameters
    ----------
        kernel : str
            The kernel to be diagonalized and plotted
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """
    X = self.properties[kernel]
    Xvp,XV = np.linalg.eigh(X)
    ncols = len(X)

    plotter = pyvista.Plotter()
    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
    def create_mesh_diagonalized_kernel(value):
        vector_number = int(round(value))
        self.plot_vector(plotter,XV[:,vector_number-1],opacity=opacity,factor=factor)
        plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(Xvp[vector_number-1]),name="eigenvalue")


    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_diagonalized_kernel, [1, len(XV)],value=1,title="Eigenvector", fmt="%1.0f")
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return


def plot_isodensity(self,plotted_isodensity="cube",opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_isodensity(plotted_isodensity="cube",opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot an isodensity

    Parameters
    ----------
        plotted_isodensity : str, optional
            The isodensity to be plotted. By default "cube"
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """
    plotter = pyvista.Plotter()
    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius,show_bonds=True)
    def create_mesh_cube(value):
        plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],self.properties["cube"],opacity=opacity,factor=factor,cutoff=value)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_cube, [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)

    return




def plot_multiple_isodensities(base_name_file,list_files,plotted_isodensity="cube",delimiter="&",opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_multiple_isodensities(range_files,plotted_isodensity="cube",delimiter=" ",opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot multiple isodensities, each one can be selected using a slider. The delimiter is replaced in the name file by number in the range_file to load multiple files
    Only one delimiter should be used in the base_name_file

    Parameters
    ----------
        base_name_file : str
            The base name of the files. If one wants to view files called H2CO.mo0a.cube, H2CO.mo1a.cube ...
            One should set the base_name_file as H2CO.mo&a.cube
        list_files : tuple
            The list of strings to replace the delimiter with
        plotted_isodensity : str, optional
            The isodensity to be plotted. By default "cube"
        delimiter : str, optional
            The
        opacity : float, optional
            The opacity of the plot. By default equals to 1
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default equals to 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()

    def open_isodensity(value,cutoff):
        value = int(round(value))
        splitted_name = base_name_file.split(delimiter)
        file = splitted_name[0] + str(value) + splitted_name[1]
        mol = rgmol.extract_cube.extract(file,do_find_bonds=1)

        if with_radius:
            mol.plot(plotter,factor=factor_radius,opacity=opacity_radius,show_bonds=True)
        plot_cube(plotter,mol.properties["voxel_origin"],mol.properties["voxel_matrix"],mol.properties["cube"],opacity=opacity,factor=factor,cutoff=cutoff)

    slider_function = _slider(open_isodensity,1,cutoff)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(lambda value: slider_function("AO",value), [int(list_files[0]),len(list_files)],value=1,title="Number", fmt="%1.0f")
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)
    plotter.show(full_screen=False)
    return




def plot_AO(self,grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_AO(grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the Atomic Orbitals of a molecule
    The Atomic Orbitals will be calculated on the grid that will be defined by the number of grid points and around the molecule. The delta defines the length to be added to the extremities of the position of the atoms.
    The order of the Atomic Orbitals is defined in the molden file

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()

    def create_mesh_AO(value,cutoff):
        plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],self.properties["AO_calculated"][int(round(value))-1],opacity=opacity,factor=factor,cutoff=cutoff)
        AO_number = int(round(value))

    if not "AO_calculated" in self.properties:
        rgmol.molecular_calculations.calculate_AO(self,grid_points,delta=delta)



    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    slider_function = _slider(create_mesh_AO,1,cutoff)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(self.properties["AO_calculated"])],value=1,title="Number", fmt="%1.0f")
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)
    return





def plot_MO(self,grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_MO(grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the Molecular Orbitals of a molecule
    The Molecular Orbitals will be calculated on the grid that will be defined by the number of grid points and around the molecule. The delta defines the length to be added to the extremities of the position of the atoms.

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter()

    if not "MO_calculated" in self.properties:
        self.properties["MO_calculated"] = [[] for k in range(len(self.properties["MO_list"]))]

    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    def create_mesh_MO(value,cutoff):
        MO_number = int(round(value))
        MO_calculated = self.calculate_MO_chosen(MO_number-1,grid_points,delta=delta)

        plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],MO_calculated,opacity=opacity,factor=factor,cutoff=cutoff)
        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["MO_energy"][MO_number-1]),name="mo energy")
        print_occupancy(plotter,self.properties["MO_occupancy"],MO_number)

    slider_function = _slider(create_mesh_MO,1,cutoff)


    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(self.properties["MO_calculated"])],value=1,title="Number", fmt="%1.0f")
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)
    return



def plot_product_MO(self,grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_product_MO(grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))

    Plot the Molecular Orbitals of a molecule on the left and the right.
    The Molecular Orbitals will be calculated on the grid that will be defined by the number of grid points and around the molecule.
    The delta defines the length to be added to the extremities of the position of the atoms.
    The center part represents the product of both sides which is calculated when pushing the button.

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        cutoff : float, optional
            The cutoff of the isodensity plot. By default .2
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function
    """

    plotter = pyvista.Plotter(shape=(1,3),border=True)

    if not "MO_calculated" in self.properties:
        self.properties["MO_calculated"] = [[] for k in range(len(self.properties["MO_list"]))]

    if with_radius:
        plotter.subplot(0,0)
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
        plotter.subplot(0,1)
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)
        plotter.subplot(0,2)
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    self.properties["MO_number_1"] = 1
    self.properties["MO_number_2"] = 1


    def create_mesh_MO_1(value):
        plotter.subplot(0,0)
        MO_number_1 = int(round(value))
        self.properties["MO_number_1"] = MO_number_1
        MO_calculated = self.calculate_MO_chosen(MO_number_1-1,grid_points,delta=delta)
        plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],MO_calculated,opacity=opacity,factor=factor,cutoff=cutoff,add_name="1")
        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["MO_energy"][MO_number_1-1]),name="mo energy")
        print_occupancy(plotter,self.properties["MO_occupancy"],MO_number_1)


    def create_mesh_MO_2(value):
        plotter.subplot(0,2)
        MO_number_2 = int(round(value))
        self.properties["MO_number_2"] = MO_number_2
        MO_calculated = self.calculate_MO_chosen(MO_number_2-1,grid_points,delta=delta)
        plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],MO_calculated,opacity=opacity,factor=factor,cutoff=cutoff,add_name="2")
        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["MO_energy"][MO_number_2-1]),name="mo energy")
        print_occupancy(plotter,self.properties["MO_occupancy"],MO_number_2)

    def calculate_product_MO(MO_ind_1,MO_ind_2):
        MO_1 = self.calculate_MO_chosen(MO_ind_1,grid_points,delta=delta)
        MO_2 = self.calculate_MO_chosen(MO_ind_2,grid_points,delta=delta)
        MO_prod = MO_1 * MO_2
        plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],MO_prod,opacity=opacity,factor=factor,cutoff=cutoff,add_name="prod")
        plotter.add_text(text=r"Product of {} and {}".format(MO_ind_1,MO_ind_2),name="mo prod")

    def button_product_MO(value):
        plotter.subplot(0,1)
        calculate_product_MO(self.properties["MO_number_1"],self.properties["MO_number_2"])




    plotter.subplot(0,0)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_MO_1, [1, len(self.properties["MO_calculated"])],value=1,title="Number", fmt="%1.0f")

    plotter.subplot(0,1)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_checkbox_button_widget(button_product_MO,size=80,color_off="blue")
    plotter.add_text(text="Calculate the product",name="calculate",position=(90.,20.))

    plotter.subplot(0,2)
    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)
    plotter.add_slider_widget(create_mesh_MO_2, [1, len(self.properties["MO_calculated"])],value=1,title="Number", fmt="%1.0f")

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.show(full_screen=False)
    return



def plot_transition_density(self,grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_transition_density(grid_points=(40,40,40),delta=3,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))


    Plot the Transition Densities of a molecule.
    All the AO and the MO will be calculated on the grid if they were not calculated.
    The grid is defined by the number of grid points and around the molecule. The delta defines the length to be added to the extremities of the position of the atoms.

    Parameters
    ----------
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        cutoff : float, optional
            The cutoff of the isodensity plot. By default .2
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        cutoff : float, optional
            The initial cutoff of the isodensity plot. By default .2
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function

    Notes
    -----
    The transition densities are defined as :

    | :math:`\\rho_0^k = \\sum_i c_i (Occ_i(r) * Virt_i(r))`
    | With the sum being on all the transitions of the excitation, :math:`Occ_i(r)` and :math:`Virt_i(r)` being respectively the occupied and the virtual molecular orbitals considered in the transition, and :math:`c_i` the coefficient of the transition.

    Examples
    --------
    For a TD-DFT calculation output :

    STATE  1:  E=   0.148431 au      4.039 eV
       | 18a ->  20a  :     0.5000 (c= 0.7071)
       | 19a ->  21a  :     0.5000 (c= -0.7071)

    The transition density will be :
        | :math:`\\rho_0^1 =  c_1 (\\psi_{18}(r) * \\psi_{20}(r)) + c_2 (\\psi_{19}(r) * \\psi_{21}(r))`
        | with :math:`c_1 = 0.7071` and :math:`c_2 = -0.7071`
    """

    plotter = pyvista.Plotter()

    #Initialize MO list
    if not "MO_calculated" in self.properties:
        self.properties["MO_calculated"] = [[] for k in range(len(self.properties["MO_list"]))]

    #Initialize transition density list
    if not "transition_density_list" in self.properties:
        self.properties["transition_density_list"] = [[] for k in range(len(self.properties["transition_list"]))]

    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)


    def create_mesh_transition_density(value,cutoff):
        transition_number = int(round(value))
        transition_density_calculated = self.calculate_chosen_transition_density(transition_number-1,grid_points,delta=delta)

        plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],transition_density_calculated,opacity=opacity,factor=factor,cutoff=cutoff)

        plotter.add_text(text=r"Energy = "+'{:3.3f} (a.u.)'.format(self.properties["transition_energy"][transition_number-1]),name="transition energy")

    slider_function = _slider(create_mesh_transition_density,1,cutoff)


    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(self.properties["transition_density_list"])],value=1,title="Number", fmt="%1.0f")
    plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))
    plotter.show(full_screen=False)





def plot_diagonalized_kernel(self,kernel,method="only eigenmodes",plotting_method="isodensity",number_eigenvectors=20,grid_points=(20,20,20),delta=3,number_isodensities=10 ,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000)):
    """
    plot_diagonalized_kernel(kernel,method="only eigenmodes",plotting_method="isodensity",number_eigenvectors=20,grid_points=(20,20,20),delta=3,number_isodensities=10 ,opacity=0.5,factor=1,with_radius=True,opacity_radius=1,factor_radius=.3,cutoff=.2,screenshot_button=True,window_size_screenshot=(1000,1000))


    Calculate and diagonalize a kernel. Only the linear response function is implemented for now.
    Only the first number_eigenvectors will be computed to limit the calculation time.
    The grid is defined by the number of grid points and around the molecule. The delta defines the length to be added to the extremities of the position of the atoms.

    Parameters
    ----------
        kernel : str
            The kernel to be diagonalized and plotted
        method : str, optional
            The method of calculation of the kernel, only eigenmodes by default. More information can be found in the notes.
        plotting_method : str, optional
            The method used for plotting. By default isodensity. The other methods are : "multiple isodensities", "volume"
        number_eigenvectors : int, optional
            Only used in the methods "total" and "partial". The number of eigenvectors to be computed
        grid_points : list of 3, optional
            The number of points for the grid in each dimension. By default (40,40,40)
        delta : float, optional
            The length added on all directions of the box containing all atomic centers. By default 3
        cutoff : float, optional
            The initial cutoff of the isodensity plot for the isodensity plotting method. By default .2
        number_isodensities : int, optional
            The number of isodensities to be plotted if the method used is multiple isodensities
        opacity : float, optional
            The opacity of the plot. By default .5
        factor : float, optional
            The factor by which the plotted_property will be multiplied. By default 1
        with_radius : bool, optional
            Chose to show the radius and the bonds between the atoms or not. By default True
        opacity_radius : float, optional
            The opacity of the radius plot. By default .8
        factor_radius : float, optional
            The factor by which the radius will be multiplied. By default .3
        screenshot_button : bool, optional
            Adds a screenshot button. True by default
        window_size_screenshot : tuple, optional
            The size of the screenshots. By default (1000,1000)

    Returns
    -------
        None
            The plotter should display when using this function


    Notes
    -----
        Because the kernels are 6-dimensional, they scale up drastically in terms of memory used.
        If one only wants to look at the eigenmodes, the "only eigenmodes" method is just that. It computes the eigenmodes without computing the total kernel. More info can be found :doc:`here<../orbitals/calculate_eigenmodes_linear_response_function>`.
        Otherwise, the partial method has been implemented which allows to remove the part of the space where the transition densities are almost zero. For each transition density, the space is sorted and the lower dense part that makes up to less than 1% is removed. In practice this removes as much as 90% of the space. More details on this method can be found :doc:`here<../orbitals/diagonalize_kernel>`.
        The last method is "total" which is just the calculation of the linear response on the whole space
    """

    if kernel != "linear_response_function":
        raise ValueError("Only linear response function implemented for now")

    if method == "only eigenmodes":
        self.calculate_eigenmodes_linear_response_function(grid_points,delta=delta)
        contrib_eigenvectors = self.properties["contibution_linear_response_eigenvectors"]


    else:
        self.diagonalize_kernel(kernel,number_eigenvectors,grid_points,method=method,delta=delta)

    eigenvectors = self.properties["linear_response_eigenvectors"]
    eigenvalues = self.properties["linear_response_eigenvalues"]
    transition_list = self.properties["transition_list"]
    transition_factor_list = self.properties["transition_factor_list"]
    plotter = pyvista.Plotter()
    if with_radius:
        self.plot(plotter,factor=factor_radius,opacity=opacity_radius)

    if plotting_method == "isodensity":

        def create_mesh_diagonalized_kernel(value,cutoff):
            vector_number = int(round(value))
            plot_cube(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],eigenvectors[vector_number-1],opacity=opacity,factor=factor,cutoff=cutoff)
            plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue")

            if method == "only eigenmodes":
                print_contribution_transition_density(plotter,vector_number,contrib_eigenvectors,transition_list,transition_factor_list)


        slider_function = _slider(create_mesh_diagonalized_kernel,1,cutoff)
        plotter.add_slider_widget(lambda value: slider_function("cutoff",value), [1e-6,1-1e-6],value=cutoff,title="Cutoff", fmt="%1.2f",pointa=(0.1,.9),pointb=(0.35,.9))

    elif plotting_method == "multiple isodensities":
        def create_mesh_diagonalized_kernel(value,cutoff):
            vector_number = int(round(value))
            plot_cube_multiple_isodensities(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],eigenvectors[vector_number-1],factor=factor)
            plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue")

            if method == "only eigenmodes":
                print_contribution_transition_density(plotter,vector_number,contrib_eigenvectors,transition_list,transition_factor_list)

    elif plotting_method == "volume":
        def create_mesh_diagonalized_kernel(value,cutoff):
            vector_number = int(round(value))
            plot_cube_volume(plotter,self.properties["voxel_origin"],self.properties["voxel_matrix"],eigenvectors[vector_number-1],factor=factor)
            plotter.add_text(text=r"eigenvalue = "+'{:3.3f} (a.u.)'.format(eigenvalues[vector_number-1]),name="eigenvalue")

            if method == "only eigenmodes":
                print_contribution_transition_density(plotter,vector_number,contrib_eigenvectors,transition_list,transition_factor_list)
    slider_function = _slider(create_mesh_diagonalized_kernel,1,cutoff)

    light = pyvista.Light((0,1,0),(0,0,0),"white",light_type="camera light",attenuation_values=(0,0,0))
    plotter.add_light(light)

    if screenshot_button:
        add_screenshot_button(plotter,window_size_screenshot)

    plotter.add_slider_widget(lambda value: slider_function("AO",value), [1, len(eigenvectors)],value=1,title="Eigenvector", fmt="%1.0f")
    plotter.show(full_screen=False)



molecule.plot = plot
molecule.plot_vector = plot_vector
molecule.plot_radius = plot_radius
molecule.plot_property = plot_property
molecule.plot_diagonalized_condensed_kernel = plot_diagonalized_condensed_kernel
molecule.plot_isodensity = plot_isodensity
molecule.plot_multiple_isodensities = plot_multiple_isodensities
molecule.plot_AO = plot_AO
molecule.plot_MO = plot_MO
molecule.plot_product_MO = plot_product_MO
molecule.plot_transition_density = plot_transition_density
molecule.plot_diagonalized_kernel = plot_diagonalized_kernel




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script contains the functions used to extract data from ADF files.
"""


import codecs
import numpy as np
from rgmol.objects import *


##########################
## Extraction functions ##
##########################

def _is_CDFT(file):
    """
    Finds if the file contains CDFT
    """
    for line in codecs.open(file,"r",encoding="utf-8"):
        if  "C O N C E P T U A L  D F T" in line:
            return True
    return False



def _rename(file):
    """
    Rename the file in order to have the .out and the .run files
    """
    if ".out" in file:
        return file,file[:-3]+"run",file[:-4]
    else: return file+".out", file+".run",file



def extract_global_descriptors(file):
    """
    Extracts the global descriptors from an adf output

    Parameters
    ----------
        file : str

    Returns
    -------
        mu      (float)    electronic chemical potential
        mu+     (float)
        mu-     (float)
        chi     (float)    electronegativity (-mu)
        eta     (float)    Hardness
        S       (float)    Softness (1/eta)
        gamma   (float)    Hyperhardness
        w       (float)    Electrophilicity index
        DEn     (float)    Dissociation energy (nucleofuge) (eV)
        DEe     (float)    Dissociation energy (electrofuge) (eV)
        w-      (float)    Electrodonating power
        w+      (float)    Electroaccepting power
        NE      (float)    Net Electrophilicity
        List of the name of the descriptors
    """
    flag,flag2=0,0
    global_desc_dict={}
    L=[]
    Name=[]
    global_desc=["mu","mu+","mu-","chi","eta","S","gamma","w","DEn","DEe","w-","w+","NE"]
    for line in codecs.open(file, 'r',encoding="utf-8"):
        #Collect data for the global descriptors. Uses flags to check when the descriptors are reached
        if "GLOBAL DESCRIPTORS" in line:
            flag=1
        elif flag==1 and "." in line:
            flag=2
            a=line.split()
            if a[-1]=="(eV)":
                L.append(float(a[-2]))
                Name.append(a[:-3])
            else:
                L.append(float(a[-1]))
                Name.append(a[:-2])
        elif flag==2 and len(line)>3:
            a=line.split()
            if a[-1]=="(eV)":
                L.append(float(a[-2]))
                Name.append(a[:-3])
            else:
                if "*" in a[-1]: L.append(None)
                else: L.append(float(a[-1]))
                Name.append(a[:-2])
        elif flag==2: break

    if len(L)==0:
        raise ImportError("The file does not contain the global descriptors")
    if len(L) != len(global_desc):
        raise ImportError("The file contains more or less descriptors than expected. ADF update ?")

    for desc in range(len(L)):
        global_desc_dict[global_desc[desc]] = L[desc]

    return L,global_desc_dict,Name

def extract_condensed_kernel(file):
    """
    Extracts the condensed kernels from an adf output

    Parameters
    ----------
        file : str

    Returns
    -------
        Linear response : ndarray
            The condensed linear response kernel
        Softness Kernel : ndarray
            The condensed softness kernel
        Name : list of str
            The name of the atoms
    """
    flag,flag2=0,0
    flag_long,flag_long2=0,0
    Lin,Sof=[],[]
    Name=[]
    for line in codecs.open(file, 'r',encoding="utf-8"):
        #Collect data for the condensed kernels. Uses flags to check when the descriptors are reached
        if "CONDENSED LINEAR RESPONSE FUNCTION" in line:
            flag=1
        elif flag==1 and "1" in line:
            flag=2
            Lin.append(line.split()[3:])
            Name.append(line.split()[:2])
        elif flag==2 and len(line)>3 and flag_long>0:
            Lin[-1]+=line.split()
            flag_long-=1
        elif flag==2 and len(line)>3:
            if int(line.split()[0])>10:
                flag_long=(int(line.split()[0])-1)//10
            Lin.append(line.split()[3:])
            Name.append(line.split()[:2])
        elif flag==2: flag=3

        if "SOFTNESS KERNEL" in line:
            flag2=1
        elif flag2==1 and "1" in line:
            flag2=2
            Sof.append(line.split()[3:])
        elif flag2==2 and len(line)>3 and flag_long2>0:
            Sof[-1]+=line.split()
            flag_long2-=1
        elif flag2==2 and len(line)>3:
            if int(line.split()[0])>10:
                flag_long2=(int(line.split()[0])-1)//10
            Sof.append(line.split()[3:])
        elif flag2==2: flag2=3
    L,S=np.zeros((len(Lin[-1]),len(Lin[-1]))),np.zeros((len(Sof[-1]),len(Sof[-1])))
    for k in range(len(Sof)):
        for j in range(k+1):

            L[j][k]=float(Lin[k][j])
            L[k][j]=float(Lin[k][j])
            S[j][k]=float(Sof[k][j])
            S[k][j]=float(Sof[k][j])
    if len(Lin)==0:
        raise ImportError("The file does not contain the kernels")

    return L,S,Name

def extract_pos(file):
    """
    Extracts the positions and the radius of the nucleus from an adf output

    Parameters
    ----------
        file : str

    Returns
    -------
        Positions : ndarray
        Radius : ndarray
        Names : list of str
    """
    flag,flag2=0,0
    Pos,Rad=[],[]
    Name=[]
    for line in codecs.open(file, 'r',encoding="utf-8"):
        #Collect data for the position and the radius. Uses flags to check when the descriptors are reached
        if "Index Symbol" in line:
            flag=1
        elif flag==1 and "1" in line:
            flag=2
            Pos.append(line.split()[2:5])
            Name.append(line.split())
        elif flag==2 and len(line)>3:
            Pos.append(line.split()[2:5])
            Name.append(line.split())
        elif flag==2: flag=3
    # print(Pos)

    return np.array(Pos,dtype="float"),Name


def extract_fukui(file):
    """
    Extracts the fukui functions from an adf output

    Parameters
    ----------
        file : str

    Returns
    -------
        f+ : ndarray
        f- : ndarray
        f0 : ndarray
        f2 : ndarray
        Name : list of str
    """
    flag,flag2=0,0
    fp,fm,f0,f2=[],[],[],[]
    Name=[]
    for line in codecs.open(file, 'r',encoding="utf-8"):
        #Collect data for the fukui function. Uses flags to check when the descriptors are reached
        if "ATOMIC DESCRIPTORS: CANONICAL ENSEMBLE" in line:
            flag=1
        elif flag==1 and "1" in line:
            flag=2
            fp.append(line.split()[3])
            fm.append(line.split()[4])
            f0.append(line.split()[5])
            f2.append(line.split()[6])
            Name.append(line.split())
        elif flag==2 and not ("--" in line):
            fp.append(line.split()[3])
            fm.append(line.split()[4])
            f0.append(line.split()[5])
            f2.append(line.split()[6])
            Name.append(line.split())

        elif flag==2: break

    return np.array(fp,dtype="float"),np.array(fm,dtype="float"),np.array(f0,dtype="float"),np.array(f2,dtype="float"),Name


def extract_bonds(file):
    """
    Extracts the bonds from an adf input

    Parameters
    ----------
        file : str

    Returns
    -------
        Bonds : list
            List of bonds
        Order_bonds : list
            List of the order of bonds
    """
    flag,flag2=0,0
    list_bonds=[]

    for line in codecs.open(file, 'r',encoding="utf-8"):
        #Collect data for the bonds. Uses flags to check when the descriptors are reached
        if "BondOrders" in line:
            flag=1
        elif flag==1 and "1" in line:
            flag=2
            list_bonds.append(line.split()[:-1]+[float(line.split()[-1])])
        elif flag==2 and not ("End" in line):
            list_bonds.append(line.split()[:-1]+[float(line.split()[-1])])
        elif flag==2: flag=3
    return list_bonds


def extract(file):
    """
    extract(file)

    Extracts all the implemented information from an adf output and input

    Parameters
    ----------
        file : str

    Returns
    -------
        mol : molecule
            molecule object
    """

    fileout,filerun,file=_rename(file)
    list_atoms=[]

    pos,Name=extract_pos(fileout)
    list_bonds=extract_bonds(filerun)
    if _is_CDFT(fileout):
        L,global_desc_dict,Name=extract_global_descriptors(fileout)
        X,S,Name=extract_condensed_kernel(fileout)
        fp,fm,f0,f2,Name=extract_fukui(fileout)
        for prop in zip(Name,pos,X,S,fp,fm,f0,f2):
            dict_properties = {"condensed linear response":prop[2],"softness kernel":prop[3],"fukui plus":prop[4],"fukui minus":prop[5],"fukui":prop[6],"dual":prop[7]}
            atom_x = atom(prop[0][1],prop[1],properties=dict_properties,nickname=prop[0][0]+prop[0][1])
            list_atoms.append(atom_x)

        global_desc_dict["condensed linear response"]=X
        global_desc_dict["softness kernel"]=S
        global_desc_dict["fukui"]=f0
        global_desc_dict["dual"]=f2

        mol = molecule(list_atoms,list_bonds,file=file,properties=global_desc_dict)

    else:
        for prop in zip(Name,pos):
            atom_x = atom(prop[0],prop[1])
            list_atoms.append(atom_x)
        mol = molecule(list_atoms,list_bonds)


    return mol




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Notes
-----

This script adds functions, and methods to the molecule objects.
These methods allow the plotting of chemical properties using pyvista.
"""

import numpy as np
import pyvista
from rgmol.objects import *
import rgmol.molecular_calculations
import os



##############################
## Representation functions ##
##############################


def Rz(alpha):
    """3D Rotation matrix around the z axis"""
    return np.array([[np.cos(alpha),-np.sin(alpha),0],[np.sin(alpha),np.cos(alpha),0],[0,0,1]])
def Ry(beta):
    """3D Rotation matrix around the y axis"""
    return np.array([[np.cos(beta),0,np.sin(beta)],[0,1,0],[-np.sin(beta),0,np.cos(beta)]])
def Rx(gamma):
    """3D Rotation matrix around the x axis"""
    return np.array([[1,0,0],[0,np.cos(gamma),-np.sin(gamma)],[0,np.sin(gamma),np.cos(gamma)]])

def corr_angle(angle,x,y):
    """Corrects the angle given by arctan to be the proper angle."""
    if x>0 and y<0:
        return -angle
    if x<0 and y>0:
        return np.pi-angle
    if x<0 and y<0:
        return np.pi+angle
    return angle



def rota_bonds(Vec,x,y,z):
    """
    Roates the bonds
    """
    #Gets the two angles alpha and beta of the vector
    Vec=Vec/np.linalg.norm(Vec)
    if Vec[0] == 0:
        if Vec[0]>0:
            alpha = np.pi/2
        else:
            alpha = -np.pi/2
    else:
        alpha=np.arctan(np.abs(Vec[1]/Vec[0]))
    alpha=corr_angle(alpha,Vec[0],Vec[1])
    Vec2=Rz(-alpha).dot(Vec)
    if Vec2[2] == 0:
        if Vec2[0]>0:
            beta = np.pi/2
        else: beta = -np.pi/2
    else:
        beta=np.arctan(abs(Vec2[0]/Vec2[2]))
    beta=corr_angle(beta,Vec2[2],Vec2[0])
    Rota=Rz(alpha).dot(Ry(beta))#The rotation matrix to convert to a vector of the x axis to a colinear vector of Vec the starting vector
    #Rotates the bonds
    x2,y2,z2=[],[],[]
    for k in range(len(x)):
        xt,yt,zt=[],[],[]
        for j in range(len(x[0])):
            pos3=Rota.dot(np.array([x[k][j],y[k][j],z[k][j]]))
            xt.append(pos3[0])
            yt.append(pos3[1])
            zt.append(pos3[2])
        x2.append(xt)
        y2.append(yt)
        z2.append(zt)
    return np.array(x2),np.array(y2),np.array(z2)


def find_angle(Vec1,Vec2):
    """
    Finds the angle between two vectors
    """
    cross=np.cross(Vec1,Vec2) #Calculate the sign of the angle
    cross= cross*(abs(cross)>1e-5)#Remove almost zero components
    scross=np.sign(np.prod(cross[np.where(cross!=0)]))
    cosangle=(Vec1.dot(Vec2) / np.linalg.norm(Vec1)/np.linalg.norm(Vec2))#cos angle
    if cosangle>1: cosangle=1 #If the two vectors are colinear, in some cases the scalar product gives 1 + 1e-15
    if cosangle<-1: cosangle=-1
    return ( np.arccos(cosangle))*scross

def rota_mol(Pos,Vec1,Vec2):
    """
    Roates the molecule
    """

    #Find the angles of the vector we want the other vector to be colinear to
    if Vec1[0]==0:
        alpha1=np.pi/2
    else: alpha1=np.arctan(np.abs(Vec1[1]/Vec1[0]))
    alpha1=corr_angle(alpha1,Vec1[0],Vec1[1])
    Vec12=Rz(-alpha1).dot(Vec1)
    if Vec12[2]==0:
        beta1=np.pi/2
    else: beta1=np.arctan(abs(Vec12[0]/Vec12[2]))
    beta1=corr_angle(beta1,Vec12[2],Vec12[0])
    Vec13=Ry(-beta1).dot(Vec12)

    if Vec2[0]==0:
        alpha2=np.pi/2
    else: alpha2=np.arctan(np.abs(Vec2[1]/Vec2[0]))
    alpha2=corr_angle(alpha2,Vec2[0],Vec2[1])
    Vec22=Rz(-alpha2).dot(Vec2)
    if Vec22[2]==0:
        beta2=np.pi/2
    else: beta2=np.arctan(abs(Vec22[0]/Vec22[2]))
    beta2=corr_angle(beta2,Vec22[2],Vec22[0])
    Vec23=Ry(-beta2).dot(Vec22)

    #V2= Rz.Ry.v2 => v2= RyT.RzT.V2
    #=> v = Rz1.Ry1.RyT.RzT.V2
    Rota=Rz(alpha1).dot(Ry(beta1).dot(Ry(-beta2).dot(Rz(-alpha2))))
    Rota_ax=Rz(alpha1).dot(Ry(beta1))
    x2,y2,z2=[],[],[]#Rotates the bonds
    # for k in range(len(Pos)):
    #     Pos[k]=Rota.dot(Pos[k])
    Pos=Rota.dot(Pos.transpose()).transpose()
    return Pos,Rota_ax



def orthonormal_basis(Pos,B,k):
    """
    Creates an orthonormal basis centered on an atom with the "z" axis perpendicular to the surface created by the 3 atoms and the "y" axis perpendicular to the bond but on the same surface created by the 3 atoms

    Inputs :    Pos (ndarray)
                B (list)
                k (int)

    Output :
                z (ndarray dim(3)) z axis
                y (ndarray dim(3)) y axis
    """

    ind=int(B[k][1])-1
    ind2=int(B[k][0])-1

    Vec=(Pos[ind2]-Pos[ind])/np.linalg.norm((Pos[ind]-Pos[ind2]))

    #In order to create an orthonormal basis, the strategy is : taking two vectors linked to two adjacent atoms that are not colinear, one can claculate the cross product wihch gives a perpendicular vector. And by taking the cross product with the last vector and one of the two first, the result is an orthonormal basis that is centered on the atom of interest
    for j in B:

        if int(j[0])-1==ind and int(j[1])-1!=ind2:

            num=int(j[1])-1
            Dist=(Pos[ind]-Pos[num])/np.linalg.norm((Pos[ind]-Pos[num]))
            angl=find_angle(Dist,Vec)

            if angl<0: Dist=-Dist
            if abs(Dist.dot(Vec))<0.95:
                per=np.cross(Vec,Dist)
                return per,np.cross(Vec,per)

        if int(j[1])-1==ind and int(j[0])-1!=ind2:

            num=int(j[0])-1
            Dist=(Pos[ind]-Pos[num])/np.linalg.norm((Pos[ind]-Pos[num]))
            angl=find_angle(Dist,Vec)

            if angl<0: Dist=-Dist
            if abs(Dist.dot(Vec))<0.95:
                per=np.cross(Vec,Dist)
                return per,np.cross(Vec,per)


        if int(j[0])-1==ind2 and int(j[1])-1!=ind:

            num=int(j[1])-1
            Dist=(Pos[ind2]-Pos[num])/np.linalg.norm((Pos[ind2]-Pos[num]))
            angl=find_angle(Dist,Vec)

            if angl<0: Dist=-Dist
            if abs(Dist.dot(Vec))<0.95:
                per=np.cross(Vec,Dist)
                return per,np.cross(Vec,per)

        if int(j[1])-1==ind2 and int(j[0])-1!=ind:

            num=int(j[0])-1
            Dist=(Pos[ind2]-Pos[num])/np.linalg.norm((Pos[ind2]-Pos[num]))
            angl=find_angle(Dist,Vec)

            if angl<0: Dist=-Dist
            if abs(Dist.dot(Vec))<0.95:
                per=np.cross(Vec,Dist)
                return per,np.cross(Vec,per)

    #Linear => we take a random vector that is not colinear
    if Vec[1]!=0 or Vec[2]!=0:
        aVec=np.copy(Vec+np.array([1,0,0]))
    else: aVec=np.copy(Vec+np.array([0,0,1]))
    aVec/=np.linalg.norm(aVec)
    per=np.cross(Vec,aVec)
    return per,np.array([1,-1,-1.3])#linear



def bonds_plotting(plotter,bonds,Pos,Vec,factor=1):
    """
    Plot the bonds
    """
    #initial values for the parameters of the bonds
    Radbond=0.05
    u=np.linspace(0,2*np.pi,30)#base of the cylinder
    v=np.linspace(0,np.pi,2) #height of the cylinder

    for k in range(len(bonds)):
        one,two=int(bonds[k][0])-1,int(bonds[k][1])-1
        order = bonds[k][2]
        Vect=Pos[one]-Pos[two]

        dist=np.linalg.norm(Pos[one]-Pos[two])

        x=Radbond*(np.outer(np.cos(u),np.ones(np.size(v))))
        y=Radbond*(np.outer(np.sin(u),np.ones(np.size(v))))
        z=(np.outer(np.ones(np.size(u)),np.linspace((abs(Vec[two]*factor)-1/20),(dist-abs(Vec[one]*factor)+1/20),np.size(v))))
        x,y,z=rota_bonds(Vect,x,y,z)
        x,y,z=x+Pos[two][0],y+Pos[two][1],z+Pos[two][2]

        if order==1:
            grid = pyvista.StructuredGrid(x,y,z)
            plotter.add_mesh(grid,name="bond_{}_{}".format(one,two),color="gray",pbr=True,roughness=.2,metallic=.7)


        elif order==1.5:
            zpe,pe=orthonormal_basis(Pos,bonds,k)#Get a orthonormal vector in order to distance the two cylinders
            pe=pe/np.linalg.norm(pe)/15
            grid = pyvista.StructuredGrid(x-pe[0],y-pe[1],z-pe[2])
            grid2 = pyvista.StructuredGrid(x+pe[0],y+pe[1],z+pe[2])
            plotter.add_mesh(grid,name="bond_{}_{}_1".format(one,two),color="gray",pbr=True,roughness=.2,metallic=.7)
            plotter.add_mesh(grid2,name="bond_{}_{}_2".format(one,two),color="white",pbr=True,roughness=.2,metallic=.7,opacity=.5)



        elif order==2:
            zpe,pe=orthonormal_basis(Pos,bonds,k)
            pe=pe/np.linalg.norm(pe)/15
            grid = pyvista.StructuredGrid(x-pe[0],y-pe[1],z-pe[2])
            grid2 = pyvista.StructuredGrid(x+pe[0],y+pe[1],z+pe[2])
            plotter.add_mesh(grid,name="bond_{}_{}_1".format(one,two),color="gray",pbr=True,roughness=.2,metallic=.7)
            plotter.add_mesh(grid2,name="bond_{}_{}_2".format(one,two),color="gray",pbr=True,roughness=.2,metallic=.7)
        else:
            zpe,pe=orthonormal_basis(Pos,bonds,k)
            pe=pe/np.linalg.norm(pe)/12
            grid = pyvista.StructuredGrid(x,y,z)
            grid2 = pyvista.StructuredGrid(x-pe[0],y-pe[1],z-pe[2])
            grid3 = pyvista.StructuredGrid(x+pe[0],y+pe[1],z+pe[2])
            plotter.add_mesh(grid,name="bond_{}_{}_1".format(one,two),color="gray",pbr=True,roughness=.2,metallic=.7)
            plotter.add_mesh(grid2,name="bond_{}_{}_2".format(one,two),color="gray",pbr=True,roughness=.2,metallic=.7)
            plotter.add_mesh(grid3,name="bond_{}_{}_3".format(one,two),color="gray",pbr=True,roughness=.2,metallic=.7)
    return





def plot_atom(plotter,atom,plotted_property="radius",opacity=1,factor=1):
    """
    plot atom as a sphere
    """
    Norm = atom.properties[plotted_property]
    atom_sphere = pyvista.Sphere(radius=Norm*factor, phi_resolution=100, theta_resolution=100,center=atom.pos)
    plotter.add_mesh(atom_sphere,color=atom.color,pbr=False,roughness=0.0,metallic=0.0,diffuse=1,opacity=opacity,name="atom_{}".format(atom.nickname))
    return


def plot_vector_atom(plotter,atom,vector,opacity=1,factor=1):
    """
    plot a vector as a sphere around an atom
    """

    colors=[[255,0,0],[255,255,255]]
    atom_sphere = pyvista.Sphere(radius=abs(vector)*factor, phi_resolution=100, theta_resolution=100,center=atom.pos)
    plotter.add_mesh(atom_sphere,name="atom_"+atom.nickname+"_vector",color=colors[(vector>0)*1],pbr=True,roughness=.4,metallic=.4,diffuse=1,opacity=opacity)

    return


def plot_cube(plotter,voxel_origin,voxel_matrix,cube,cutoff=0.1,opacity=1,factor=1,add_name=""):
    """
    Plots an isodensity
    """

    nx,ny,nz = np.shape(cube)
    cube_transposed = np.transpose(cube,(2,1,0))

    grid = pyvista.ImageData(dimensions=(nx,ny,nz),spacing=(voxel_matrix[0][0], voxel_matrix[1][1], voxel_matrix[2][2]),origin=voxel_origin)

    #Calculate cube density
    # cube_density = cube_transposed**2 * voxel_matrix[0][0] * voxel_matrix[1][1] * voxel_matrix[2][2]
    cube_density = abs(cube_transposed) * voxel_matrix[0][0] * voxel_matrix[1][1] * voxel_matrix[2][2]

    #Calculate renormalization as for some reason some cube files are not normalized
    cube_density = cube_density / np.sum(cube_density)

    array_sort = np.argsort(cube_density,axis=None)[::-1]
    cube_sorted = cube_density.flatten()[array_sort]
    cube_values_sorted = np.cumsum(cube_sorted)

    #Find how to unsort the array. There should be a more efficient way to do this
    indexes = np.arange(len(array_sort),dtype=int)
    array_unsort = np.zeros(len(array_sort),dtype=int)

    for k in range(len(array_sort)):
        array_unsort[array_sort[k]] = indexes[k]
    cube_values = cube_values_sorted[array_unsort]

    cube_values_positive = cube_values + (cube_transposed<0).flatten() * (1-cube_values)
    cube_values_negative = cube_values + (cube_transposed>0).flatten() * (1-cube_values)

    contour_positive = grid.contour(isosurfaces=2,scalars=cube_values_positive,rng=[0,1-cutoff])
    contour_negative = grid.contour(isosurfaces=2,scalars=cube_values_negative,rng=[0,1-cutoff])


    if len(contour_positive.point_data["Contour Data"]):
        plotter.add_mesh(contour_positive,name="isosurface_cube_positive"+add_name,opacity=opacity,pbr=True,roughness=.5,metallic=.2,color="red")
    else:
        plotter.remove_actor("isosurface_cube_positive"+add_name)
    if len(contour_negative.point_data["Contour Data"]):
        plotter.add_mesh(contour_negative,name="isosurface_cube_negative"+add_name,opacity=opacity,pbr=True,roughness=.5,metallic=.2,color="blue")
    else:
        plotter.remove_actor("isosurface_cube_negative"+add_name)



def plot_cube_multiple_isodensities(plotter,voxel_origin,voxel_matrix,cube,number_isodensities=10,opacity=1,factor=1,add_name=""):
    """
    plot multiple isodensities
    """

    nx,ny,nz = np.shape(cube)
    cube_transposed = np.transpose(cube,(2,1,0))

    grid = pyvista.ImageData(dimensions=(nx,ny,nz),spacing=(voxel_matrix[0][0], voxel_matrix[1][1], voxel_matrix[2][2]),origin=voxel_origin)

    #Calculate cube density
    cube_density = abs(cube_transposed) * voxel_matrix[0][0] * voxel_matrix[1][1] * voxel_matrix[2][2]

    #Calculate renormalization as for some reason some cube files are not normalized
    cube_density = cube_density / np.sum(cube_density)

    array_sort = np.argsort(cube_density,axis=None)[::-1]
    cube_sorted = cube_density.flatten()[array_sort]
    cube_values_sorted = np.cumsum(cube_sorted)

    #Find how to unsort the array. There should be a more efficient way to do this
    indexes = np.arange(len(array_sort),dtype=int)
    array_unsort = np.zeros(len(array_sort),dtype=int)

    for k in range(len(array_sort)):
        array_unsort[array_sort[k]] = indexes[k]
    cube_values = cube_values_sorted[array_unsort]

    cube_values_positive = cube_values + (cube_transposed<0).flatten() * (1-cube_values)
    cube_values_negative = cube_values + (cube_transposed>0).flatten() * (1-cube_values)

    # for cutoff in cutoffs
    contour_positive = grid.contour(isosurfaces=number_isodensities,scalars=cube_values_positive,rng=[0,0.9])
    contour_negative = grid.contour(isosurfaces=number_isodensities,scalars=cube_values_negative,rng=[0,0.9])

    import matplotlib.colors as mpc

    empty = np.ones((256,3))
    cmap_pos = mpc.ListedColormap(empty * np.array([1.,0.,0.]*np.linspace(.5,1,256).reshape((256,1))))
    cmap_neg = mpc.ListedColormap(empty * np.array([0.,0.,1.]*np.linspace(.5,1,256).reshape((256,1))))


    point_positive = contour_positive.points
    point_negative = contour_negative.points

    if len(contour_positive.point_data["Contour Data"]):

        plotter.add_mesh(contour_positive,name="isosurface_cube_positive"+add_name,opacity=1-contour_positive["Contour Data"],scalars=contour_positive["Contour Data"],pbr=True,roughness=.5,metallic=.2,cmap=cmap_pos,show_scalar_bar=False)
    else:
        plotter.remove_actor("isosurface_cube_positive"+add_name)
    if len(contour_negative.point_data["Contour Data"]):
        plotter.add_mesh(contour_negative,name="isosurface_cube_negative"+add_name,opacity=1-contour_negative["Contour Data"],scalars=contour_negative["Contour Data"],pbr=True,roughness=.5,metallic=.2,cmap=cmap_neg,show_scalar_bar=False)
    else:
        plotter.remove_actor("isosurface_cube_negative"+add_name)



def plot_cube_volume(plotter,voxel_origin,voxel_matrix,cube,opacity=1,factor=1,add_name=""):
    """
    plot the cube as a volume
    """

    nx,ny,nz = np.shape(cube)
    cube_transposed = np.transpose(cube,(2,1,0))

    coordinates_kept = np.zeros((nz,ny,nx),dtype="bool")
    coordinates_ravelled = np.arange(nx*ny*nz)

    cube_flatten = abs(cube_transposed).flatten()

    sorting_array = np.argsort(cube_flatten,axis=None)[::-1]
    norm_cube = np.sum(cube_flatten)
    cube_cumsum = np.cumsum(cube_flatten[sorting_array])

    coordinates_sorted = coordinates_ravelled[sorting_array]

    coordinates_ravelled_kept = coordinates_sorted[cube_cumsum < (norm_cube*0.99)]

    for coordinates in coordinates_ravelled_kept:
        coordinates_kept[np.unravel_index(coordinates,(nx,ny,nz))] = True

    coordinates = np.arange(nx*ny*nz)[coordinates_kept.flatten()]
    minx,miny,minz,maxx,maxy,maxz = 100,100,100,0,0,0
    for co in coordinates:
        x,y,z = np.unravel_index(co,(nx,ny,nz))
        minx = min(x,minx)
        maxx = max(x,maxx)
        miny = min(y,miny)
        maxy = max(y,maxy)
        minz = min(z,minz)
        maxz = max(z,maxz)

    new_nx,new_ny,new_nz = maxx-minx+1,maxy-miny+1,maxz-minz+1
    # print(new_nx,new_ny,new_nz)

    voxel_origin_0 = voxel_origin[0] + minx * voxel_matrix[0][0]
    voxel_origin_1 = voxel_origin[1] + miny * voxel_matrix[1][1]
    voxel_origin_2 = voxel_origin[2] + minz * voxel_matrix[2][2]

    grid = pyvista.ImageData(dimensions=(new_nx,new_ny,new_nz),spacing=(voxel_matrix[0][0], voxel_matrix[1][1], voxel_matrix[2][2]),origin=(voxel_origin_0,voxel_origin_1,voxel_origin_2))

    cube_data = cube_flatten[coordinates_kept.flatten()]

    cube = pyvista.Cube(x_length = voxel_matrix[0][0],y_length = voxel_matrix[1][1],z_length = voxel_matrix[2][2])
    glyphs = grid.glyph(orient=False, geom=cube, scale=False)

    scalars = abs((cube_transposed[minz:maxz+1,miny:maxy+1,minx:maxx+1]).reshape((new_nx,new_ny,new_nz,1)) * np.ones((1,1,1,6)))

    plotter.add_mesh(glyphs,scalars=scalars,opacity="linear",show_scalar_bar=False,name="volume"+add_name)





def print_contribution_transition_density(plotter,vector_number,contrib_eigenvectors,transition_list,transition_factor_list,divy=1):
    """
    Prints the contribution of each transition density for an eigenvector
    """


    font_size_title = int(16/divy*plotter.window_size[1]/1000)
    font_size = int(14/divy*plotter.window_size[1]/1000)
    pos_y = plotter.window_size[1]/divy-200*plotter.window_size[1]/1000

    plotter.add_text(text=r"Contribution of tranisition densities",name="contrib_name",position=(0,pos_y),font_size=font_size_title)

    array_sort = np.argsort(abs(contrib_eigenvectors[vector_number-1]))[::-1]
    contributions = contrib_eigenvectors[vector_number-1]
    contrib_sorted = contrib_eigenvectors[vector_number-1][array_sort]
    contrib_indices_sorted = np.arange(len(contrib_eigenvectors[vector_number-1]))[array_sort]


    lin_transition = np.array([transition for transitions in transition_list for transi in transitions for transition in transi])

    occ = lin_transition[::2]
    virt = lin_transition[1::2]
    max_occupied = np.max(occ)
    max_virtual= np.max(virt)
    transition_contrib = np.zeros((max_occupied+1,max_virtual+1))

    text_contrib = ""
    compt_contrib = 0
    for contrib in range(len(contrib_sorted)):
        if abs(contrib_sorted[contrib])<0.05:
            break
        text_contrib += r"C_"+"{}".format(contrib_indices_sorted[contrib]+1)+": {:3.3f}\n".format(contrib_sorted[contrib])
        compt_contrib += 1

    plotter.add_text(text=text_contrib,name="contrib",font_size=font_size,position=(20.0,pos_y-2*font_size_title-2*font_size*compt_contrib))
    plotter.add_text(text="Contribution of transitions",name="transi_title",font_size=font_size_title,position=(0.0,pos_y-3*font_size_title-2*font_size*(compt_contrib)))

    for contrib in range(len(contributions)):
        transitions = transition_list[contrib]
        transition_factors = transition_factor_list[contrib]

        for transi,transi_factor in zip(transitions,transition_factors):
            transition_contrib[transi[0],transi[1]] += contributions[contrib] * transi_factor[0]

    transition_contrib = transition_contrib / np.sum(abs(transition_contrib))
    compt = 0

    array_sort = np.argsort(abs(transition_contrib).ravel())[::-1]

    transition_contrib_sorted = (transition_contrib.ravel())[array_sort]

    text_contrib = ""
    for trans in range(len(transition_contrib_sorted)):
        occ,virt = np.unravel_index(array_sort[trans],np.shape(transition_contrib))
        if abs(transition_contrib[occ,virt]) < 0.05:
            break
        text_contrib += "{} -> {}".format(occ,virt)+": {:3.3f}\n".format(transition_contrib[occ,virt])
        compt += 1

    plotter.add_text(text=text_contrib,name="transi",font_size=font_size,position=(20.0,pos_y-4*font_size_title-2*font_size*compt_contrib-2*font_size*compt))



def print_occupancy(plotter,MO_occupancy,MO_number,divy=1):
    """
    Prints the contribution of each transition density for an eigenvector
    """

    LUMO = np.argmin(MO_occupancy)

    # plotter.add_text(text=r"Occupancy : "+'{:1.1f}'.format(MO_occupancy[MO_number-1]),name="mo occupancy",position=(plotter.window_size[0]*(divx) + 20,plotter.window_size[1]*(divy)-100))

    font_size = int(18/divy*plotter.window_size[1]/1000)
    pos_y = plotter.window_size[1]/divy-200*plotter.window_size[1]/1000

    plotter.add_text(text=r"Occupancy : "+'{:1.1f}'.format(MO_occupancy[MO_number-1]),name="mo occupancy",position=(20,pos_y),font_size=font_size)
    if MO_number-1 < LUMO:
        if MO_number-1 == LUMO-1:
            plotter.add_text(text=r"HOMO",name="mo occupancy lumo",position=(20.0,pos_y-2*font_size),font_size=font_size)
        else:
            plotter.add_text(text=r"HOMO - {}".format(LUMO-MO_number),name="mo occupancy lumo",position=(20.0,pos_y-2*font_size),font_size=font_size)
    else:
        if MO_number-1 == LUMO:
            plotter.add_text(text=r"LUMO",name="mo occupancy lumo",position=(20.0,pos_y-font_size),font_size=font_size)
        else:
            plotter.add_text(text=r"LUMO + {}".format(MO_number-1-LUMO),name="mo occupancy lumo",position=(20.0,pos_y-2*font_size),font_size=font_size)



def add_screenshot_button(plotter,window_size):
    """
    Adds a screenshot button on the plotter
    """

    def button_screenshot(value):
        for x in range(plotter.shape[0]):
            for y in range(plotter.shape[1]):
                plotter.subplot(x,y)
                for actor,content in plotter.actors.items():
                    if not "bond" in actor and not "isosurface" in actor and not "atom" in actor:
                        content.SetVisibility(False)
        #
        for slider in plotter.slider_widgets:
            repres = slider.GetRepresentation()
            repres.VisibilityOff()
            slider.SetRepresentation(repres)
        for button in plotter.button_widgets:
            repres = button.GetRepresentation()
            repres.VisibilityOff()
            button.SetRepresentation(repres)

        List_files = os.listdir()
        d = 0
        while "rgmol_screenshot_{}.png".format(d) in List_files:
            d+=1
        plotter.screenshot(filename="rgmol_screenshot_{}.png".format(d),window_size=window_size)

        for x in range(plotter.shape[0]):
            for y in range(plotter.shape[1]):
                plotter.subplot(x,y)
                for actor,content in plotter.actors.items():
                    content.SetVisibility(True)

        for slider in plotter.slider_widgets:
            repres = slider.GetRepresentation()
            repres.VisibilityOn()
            slider.SetRepresentation(repres)
        for button in plotter.button_widgets:
            repres = button.GetRepresentation()
            repres.VisibilityOn()
            button.SetRepresentation(repres)

    plotter.add_text(text="Screenshot",name="screenshot",position=(90.,20.))
    plotter.add_checkbox_button_widget(button_screenshot,size=80,color_off="blue")


